@@ -6,7 +6,7 @@
  * 	Gonzalo Paniagua Javier
  *
  * Copyright (c) 2002 Daniel Lopez Ridruejo
- *           (c) 2002-2005 Novell, Inc.
+ *           (c) 2002-2006 Novell, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -82,6 +82,11 @@ typedef struct {
 	char auto_app_set;
 } module_cfg;
 
+typedef struct {
+	uint32_t client_block_buffer_size;
+	char *client_block_buffer;
+} request_data;
+
 /* */
 static int
 search_for_alias (const char *alias, module_cfg *config)
@@ -129,6 +134,7 @@ set_auto_application (cmd_parms *cmd, void *mconfig, const char *value)
 			return apr_pstrdup (cmd->pool, "Conflicting values for MonoAutoApplication.");
 
 		sconfig->auto_app = FALSE;
+		/* TODO: Copiar de 'XXGLOBAL' a 'default' */
 	} else if (!strcasecmp (value, "enabled")) {
 		if (sconfig->auto_app_set && sconfig->auto_app != TRUE)
 			return apr_pstrdup (cmd->pool, "Conflicting values for MonoAutoApplication.");
@@ -159,8 +165,8 @@ add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_
 	server->alias = apr_pstrdup (pool, alias);
 	server->filename = NULL;
 	server->run_xsp = "True";
-	server->executable_path = EXECUTABLE_PATH;
-	server->path = MONO_PATH;
+	/* (Obsolete) server->executable_path = EXECUTABLE_PATH; */
+	server->path = NULL;
 	server->server_path = MODMONO_SERVER_PATH;
 	server->applications = NULL;
 	server->wapidir = WAPIDIR;
@@ -173,7 +179,7 @@ add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_
 	server->listen_address = NULL;
 	server->max_cpu_time = NULL;
 	server->max_memory = NULL;
-	server->debug = "False";
+	server->debug = NULL;
 	server->env_vars = NULL;
 	server->status = FORK_NONE;
 	server->is_virtual = is_virtual;
@@ -206,20 +212,43 @@ store_config_xsp (cmd_parms *cmd, void *notused, const char *first, const char *
 	offset = (unsigned long) cmd->info;
 	DEBUG_PRINT (1, "store_config %lu '%s' '%s'", offset, first, second);
 	config = ap_get_module_config (cmd->server->module_config, &mono_module);
-	if (!config->auto_app_set)
-		config->auto_app = FALSE; /* disable autoapp if there's any other application */
-
 	if (second == NULL) {
+		if (config->auto_app) {
+			idx = search_for_alias ("XXGLOBAL", config);
+			ptr = (char *) &config->servers [idx];
+			ptr += offset;
+			value = first;
+
+			/* MonoApplications/AddMonoApplications are accumulative */
+			if (offset == APR_OFFSETOF (xsp_data, applications))
+				prev_value = *((char **) ptr);
+
+			if (prev_value != NULL) {
+				new_value = apr_pstrcat (cmd->pool, prev_value, ",", value, NULL);
+			} else {
+				new_value = apr_pstrdup (cmd->pool, value);
+			}
+
+			*((char **) ptr) = new_value;
+			return NULL;
+		}
 		alias = "default";
-		if (cmd->server->is_virtual) alias = cmd->server->server_hostname;
+		if (cmd->server->is_virtual)
+			alias = cmd->server->server_hostname;
 		value = first;
 		is_default = 1;
 	} else {
+		if (!strcmp (first, "XXGLOBAL"))
+			return apr_pstrdup (cmd->pool, "XXGLOBAL is a reserved application identifier.");
 		alias = first;
 		value = second;
 		is_default = (!strcmp (alias, "default"));
 	}
 
+	/* Disable autoapp if there's any other application. MonoDebug is excluded. */
+	if (!config->auto_app_set)
+		config->auto_app = FALSE;
+
 	idx = search_for_alias (alias, config);
 	if (idx == -1)
 		idx = add_xsp_server (cmd->pool, alias, config, is_default, cmd->server->is_virtual);
@@ -329,6 +358,8 @@ connection_get_remote_port (conn_rec *c)
 { 
 #ifdef APACHE13
 	return  ntohs (c->remote_addr.sin_port);
+#elif defined(APACHE22)
+	return c->remote_addr->port;
 #else
 	apr_port_t port;
 	apr_sockaddr_port_get (&port, c->remote_addr);
@@ -342,6 +373,8 @@ connection_get_local_port (request_rec *r)
 {
 #ifdef APACHE13  
 	return ap_get_server_port (r);
+#elif defined(APACHE22)
+	return r->connection->local_addr->port;
 #else
 	apr_port_t port;
 	apr_sockaddr_port_get (&port, r->connection->local_addr);
@@ -392,7 +425,7 @@ setup_client_block (request_rec *r)
 	if (r->read_length)
 		return APR_SUCCESS;
 
-	return ap_setup_client_block (r, REQUEST_CHUNKED_ERROR);
+	return ap_setup_client_block (r, REQUEST_CHUNKED_DECHUNK);
 }
 
 static int
@@ -559,6 +592,28 @@ remove_http_vars (apr_table_t *table)
 	} while (t_elt < t_end);
 }
 
+static char *
+get_client_block_buffer (request_rec *r, uint32_t requested_size, uint32_t *actual_size)
+{
+	request_data *rd = ap_get_module_config (r->request_config, &mono_module);
+
+	if (rd == NULL)	{
+		rd = apr_pcalloc (r->pool, sizeof (request_data));
+		ap_set_module_config (r->request_config, &mono_module, rd);
+	}
+
+	if (requested_size > 1024 * 1024)
+		requested_size = 1024 * 1024;
+
+	if (requested_size > rd->client_block_buffer_size) {
+		rd->client_block_buffer = apr_pcalloc (r->pool, requested_size);
+		rd->client_block_buffer_size = requested_size;
+	}
+
+	*actual_size = requested_size;
+	return rd->client_block_buffer;
+}
+
 static int
 do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 {
@@ -566,6 +621,7 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 	char *str;
 	const char *cstr;
 	int32_t i;
+	uint32_t actual_size;
 	int status = 0;
 
 	if (command < 0 || command >= LAST_COMMAND) {
@@ -626,8 +682,8 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 			break;
 
 		i = INT_FROM_LE (i);
-		str = apr_pcalloc (r->pool, i);
-		i = ap_get_client_block (r, str, i);
+		str = get_client_block_buffer (r, (uint32_t) i, &actual_size);
+		i = ap_get_client_block (r, str, actual_size);
 		i = LE_FROM_INT (i);
 		status = write_data (sock, &i, sizeof (int32_t));
 		i = INT_FROM_LE (i);
@@ -650,6 +706,7 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result)
 		return FALSE;
 	case IS_CONNECTED:
 		*result = (r->connection->aborted ? 0 : 1);
+		status = write_data (sock, result, sizeof (int32_t));
 		break;
 	case MYNOT_FOUND:
 		ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
@@ -966,7 +1023,6 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	int argi;
 	char *path;
 	char *tmp;
-	char *monodir;
 	char *serverdir;
 	char *wapidir;
 	int max_memory = 0;
@@ -1054,22 +1110,15 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	if (tmp == NULL)
 		tmp = "";
 
-	monodir = get_directory (pool, config->executable_path);
-	DEBUG_PRINT (1, "monodir: %s", monodir);
 	serverdir = get_directory (pool, config->server_path);
 	DEBUG_PRINT (1, "serverdir: %s", serverdir);
-	if (strcmp (monodir, serverdir)) {
-		path = apr_pcalloc (pool, strlen (tmp) + strlen (monodir) +
-					strlen (serverdir) + 3);
-		sprintf (path, "%s:%s:%s", monodir, serverdir, tmp);
-	} else {
-		path = apr_pcalloc (pool, strlen (tmp) + strlen (monodir) + 2);
-		sprintf (path, "%s:%s", monodir, tmp);
-	}
+	path = apr_pcalloc (pool, strlen (tmp) + strlen (serverdir) + 2);
+	sprintf (path, "%s:%s", serverdir, tmp);
 
 	DEBUG_PRINT (1, "PATH after: %s", path);
 	SETENV (pool, "PATH", path);
-	SETENV (pool, "MONO_PATH", config->path);
+	if (config->path != NULL)
+		SETENV (pool, "MONO_PATH", config->path);
 	wapidir = apr_pcalloc (pool, strlen (config->wapidir) + 5 + 2);
 	sprintf (wapidir, "%s/%s", config->wapidir, ".wapi");
 	mkdir (wapidir, 0700);
@@ -1080,13 +1129,11 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	}
 
 	SETENV (pool, "MONO_SHARED_DIR", config->wapidir);
+	if (config->debug && !strcasecmp (config->debug, "True"))
+		SETENV (pool, "MONO_OPTIONS", "--debug");
 
 	memset (argv, 0, sizeof (char *) * MAXARGS);
 	argi = 0;
-	argv [argi++] = config->executable_path;
-	if (!strcasecmp (config->debug, "True"))
-		argv [argi++] = "--debug";
-
 	argv [argi++] = config->server_path;
 	if (config->listen_port != NULL) {
 		char *la;
@@ -1160,20 +1207,22 @@ static apr_status_t
 setup_socket (apr_socket_t **sock, xsp_data *conf, apr_pool_t *pool)
 {
 	apr_status_t rv;
-	int family;
+	int family, proto;
 
-	family = (conf->listen_port != NULL) ? PF_INET : PF_UNIX;
+	family = (conf->listen_port != NULL) ? AF_UNSPEC : PF_UNIX;
+	/* APR_PROTO_TCP = 6 */
+	proto = (family == AF_UNSPEC) ? 6 : 0;
 #ifdef APACHE2
-	rv = apr_socket_create (sock, family, SOCK_STREAM, pool);
+	rv = APR_SOCKET_CREATE (sock, family, SOCK_STREAM, proto, pool);
 #else
 	(*sock)->fd = ap_psocket (pool, family, SOCK_STREAM, 0);
 	(*sock)->pool = pool;
 	rv = ((*sock)->fd != -1) ? APR_SUCCESS : -1;
 #endif
 	if (rv != APR_SUCCESS) {
+		int err= errno;
 		ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
-			      "mod_mono: error creating socket.");
-
+			      "mod_mono: error creating socket: %d %s", err, strerror (err));
 		return rv;
 	}
 
@@ -1360,6 +1409,7 @@ mono_execute_request (request_rec *r, char auto_app)
 		idx = search_for_alias (NULL, config);
 
 	DEBUG_PRINT (2, "idx = %d", idx);
+
 	if (idx < 0) {
 		DEBUG_PRINT (2, "Alias not found. Checking for auto-applications.");
 		if (config->auto_app)
@@ -1461,7 +1511,7 @@ mono_handler (request_rec *r)
 {
 	module_cfg *config;
 
-	if (!strcmp (r->handler, "mono")) {
+	if (r->handler != NULL && !strcmp (r->handler, "mono")) {
 		DEBUG_PRINT (1, "handler: %s", r->handler);
 		return mono_execute_request (r, FALSE);
 	}
@@ -1724,6 +1774,7 @@ mono_child_init (
 #ifdef APACHE13
 static const handler_rec mono_handlers [] = {
 	{ "mono", mono_handler },
+	{ "application/x-asp-net", mono_handler },
 	{ "mono-ctrl", mono_control_panel_handler },
 	{ NULL, NULL }
 };
@@ -1764,7 +1815,7 @@ MAKE_CMD12 (MonoRunXSP, run_xsp,
 	),
 
 MAKE_CMD12 (MonoExecutablePath, executable_path,
-	"If MonoRunXSP is True, this is the full path where mono is located. "
+	"(Obsolete) If MonoRunXSP is True, this is the full path where mono is located. "
 	"Default: /usr/bin/mono"
 	),
 
@@ -1774,7 +1825,7 @@ MAKE_CMD12 (MonoPath, path,
 	),
 
 MAKE_CMD12 (MonoServerPath, server_path,
-	"If MonoRunXSP is True, this is the full path to mod-mono-server.exe. "
+	"If MonoRunXSP is True, this is the full path to the mod-mono-server script. "
 	"Default: " MODMONO_SERVER_PATH
 	),
 
